#!/usr/bin/env python
"""Generate uppaal model from data of model learned with learnlib """
# Author: Harco Kuppens


import sys, re, pprint  # modules from standard lib (python 2.6 and later)

"""

LTS

  read_from_file()
  write_to_file(x.aut or x.dot)

"""

class LabeledTransition(object):
   """
   labeled transition consisting of :
      - source : start state of transition
      - target : end state of transition
      - label  : label of transition
   """
   def __init__(self, source, target, label):
       self.source=source
       self.target=target
       self.label=label
   def __repr__(self):
       return "LabeledTransition(" + repr(self.source) + "," + repr(self.target) + "," + repr(self.label) + ")"
   def __str__(self):
       return str(self.source) + " -- "  +  self.getLabel() + " -->" + str(self.target)
   def getLabel(self):
       return  self.label


class IOTransition(object):
   """
   labeled transition consisting of :
      - source : start state of transition
      - target : end state of transition
      - input  : input of transition
      - output  : output of transition
   """
   def __init__(self, source, target, input,output):
       self.source=source
       self.target=target
       self.input=input
       self.output=output

   def __repr__(self):
       return "IOTransition(" + repr(self.source) + "," + repr(self.target) + "," + repr(self.input)  + "," + repr(self.output) + ")"
   def __str__(self):
       return str(self.source) + " -- " +  self.getLabel() + " --> " + str(self.target)

   def getLabel(self):
       return  self.input + "/" + self.output

class LTS(object):
   """
   Transtition system
      has only startState, states and transititions
      Note: transition can be any kind of transition implementing a getLabel() method
   """
   def __init__(self, startState=None,states=[],transitions=[]):
       self.startState=startState
       self.states=states
       self.transitions=transitions

   def __str__(self):
       strings=["startState=" + str(self.startState)]
       for t in self.transitions:
           strings.append(str(t))
       return "\n".join(strings)

   def reformatInAldeberanStyle(self):
       pass

   def getAutStr(self):
       return "aut"

   def writeAutFile(self,aut_file):
        f=file(aut_file,'w')
        f.write(self.getAutStr)
        f.close()

   def importDotFile(aut_file):
        f=file(aut_file)
def mealy2lts(mealy):
    return iots2lts(mealy2iots(mealy))


def iots2lts(iots):
    """
    Transform  IOTS machine with  separate 'input' and 'output' fields on
    each transition to a LTS machine.

    Splits transitions with both an 'input' and 'output' into two transitions

       Makes two sequential transitions with a dummy state in between:
        - dummy state <midstate> is labeled :
             m_<counter>
        - first transition :
             <source> -> <midstate>  with label=?<input>
        - second transition :
             <midstate> -> <target>  with label=!<output>
    """
    trans_out=[]
    id=len(iots.states)
    for t in iots.transitions:
        midstate= 'm' + str(id)
        trans_out.append(LabeledTransition(
            source= t.source,
            target= midstate,
            label = "?" + t.input,
        ))
        trans_out.append(LabeledTransition(
            source= midstate,
            target= t.target,
            label = "!" + t.output,
        ))
        id=id+1

    states=set()
    for t in trans_out:
        states.add(t.source)
        states.add(t.target)

    return LTS(iots.startState,states,trans_out)

def mealy2iots(mealy):
    """
    transform  Mealy machine with "input/output" labels to a IOTS machine
    having transitions with separate input and output labels.
    Thus in a IOTS each transition has an 'input' field and an 'output' field.
    """

    trans_out=[]
    for t in mealy.transitions:
        [inputstr,outputstr]=t.label.split('/')
        trans_out.append( IOTransition(
          source = t.source,
          target = t.target,
          input  = inputstr,
          output = outputstr
        ))

    return LTS(mealy.startState,mealy.states,trans_out)



def get_mealy_from_dotfile(dot_file):
    """ Get mealy transition system from graphviz dot file

        The dot file:
           - describes a mealy machine as a digraph with labels
           - label is of form :  <input>/<ouput>
           - encodes the start state with the color='red' attribute
             note: this corresponds with the highlighted state in learnlib API


        Returns: [start_state,transions]
        Where :
           - start_state: start state label
           - transitions: list of transitions

    """

    start_state='unknown'
    f=file(dot_file)
    lines=f.readlines()

    # find start state
    for line in lines:
        line=line.replace('label=""','')
        if line.find('[color="red"]') != -1:
            start_state=line[:line.find(' ')]
            break


    # get transitions
    transitions=[]
    for line in lines:
        if line.find('->') != -1:
            transitions.append(line)

    # throw away transitions with the keywords : quiescence or inconsistency or undefined
    #transitions = [ t for t in transitions if ( 'quiescence' not in t ) and ( 'inconsistency' not in t  )  and ( 'undefined' not in t  )]

    trans_out=[]
    regexpr_transition=re.compile(r'\s*(\w*)\s*-\>\s*(\w*)\s*\[label=\<(.*)\>\]')
    regexpr_tag=re.compile(r'<[^>]+>')
    for transition in transitions:
        match=regexpr_transition.match(transition)
        if match:
            match=match.groups()
            label=regexpr_tag.sub('',match[2])
            trans_out.append( LabeledTransition(
              source = match[0],
              target = match[1],
              label = label
            ))


    states=set()
    for t in trans_out:
        states.add(t.source)
        states.add(t.target)


    return LTS(start_state,states,trans_out)







def lts2aut(lts):
    """
     http://www.mcrl2.org/mcrl2/wiki/index.php/AUT_Syntax
       - labels are strings
       - states are represented by integers
       - first_state should always be 0
    """
    pass

def serialize_aut(lts):

    nr_of_transitions=len(lts.transitions)
    nr_of_states=len(lts.states)
    strings=[ "des(" + startState[1:] + ","  + str(nr_of_transitions) + "," + str(nr_of_states) +")"]
    for t in lts.transitions:
       strings.append("("+t.source[1:] + "," + '"' + t.label + '"' + "," + t.target[1:] + ")" )

    return "\n".join(strings)


def dot2aut(dot_filename_in):
    """
       from mealy machine in a .dot file written by DotUtil.write of learnlib
       we create an .aut file containing an lts where input and output each
       have its own labeled transition. An input transition has a
       label starting with '?' and an output transition has a label
       starting with '!'

    """
    m=get_mealy_from_dotfile(dot_filename_in)
    io_transitions=parse_labels_of_mealy_lts(transitions) # each transition has input and output
    [states,transitions]=split_io_transitions_in_separate_input_and_output_transition(io_transitions,len(states)) # each transition only has label again
    return transitions2aut(transitions,start_state,len(states))



if __name__ == "__main__":
    print dot2aut(*sys.argv[1:])

